from builtins import str
from builtins import range
import os

from olmonk import ConfigDataValidator as CDV
from olmonk import helpers as hlp
import pandas as pd
import numpy as np
from datum import constants as const
from datum import custom_exceptions


@custom_exceptions.handleError
def validate_input_file(path):
    """
    This is the function for checking basic validation of file.
    If any exception raises during validation check , handleError
    decorator will catch it and processed accordingly.
    Parameters
    ----------
    path : string
        path of file
    Returns
    -------
    :return: True, if no exception is raised
    """

    if not os.path.isfile(path):
        raise custom_exceptions.FileExistError

    if os.stat(path).st_size == 0:
        raise custom_exceptions.FileEmptyError

    return True


@custom_exceptions.handleError
def read_file(input_file_path):
    """This function reads the input file in xls, xlsx,
    txt and csv format
    Parameters
    ----------
    input_file_path : string
        path of input file
    Returns
    -------
    df : input file dataframe after reading the file
    """

    if os.path.splitext(input_file_path)[1] in const.EXCEL_FILES:
        df = pd.read_excel(input_file_path, header=0)

    elif os.path.splitext(input_file_path)[1] == const.CSV_FILE:
        df = pd.read_csv(input_file_path, header=0)

    elif os.path.splitext(input_file_path)[1] == const.TXT_FILE:
        pd.set_option('display.float_format',lambda x: '%f' % x)
        try:
            df = pd.read_csv(input_file_path, sep='\t|,', engine='python')
        except Exception as e:
            try:
                df = pd.read_table(input_file_path, header=0)
            except:
                df = pd.read_csv(input_file_path)

    else:
        raise custom_exceptions.FileExtensionError

    return df


def fill_missing(df):
    """
    To fill Missing values with 0 in the dataframe.

    Parameters
    ----------
    df : dataframe
        dataframe to be checked
    Returns
    -------
    dataframe filled with 0
    """
    return df.fillna(0)


def missing_data_report(missing_df, logs):
    """
    Log the information with action and message

    Parameters
    ----------
    missing_df : dataframe
        dataframe to be checked
    Returns
    -------
    Dict : logs of errors and warnings
    """

    for index, row in missing_df.iterrows():
        logs[const.WARNINGS][const.MESSAGE].append(
            const.ROW_NUMBER + ' ' + str(row[const.ROW_NUMBER_COLUMN]) + ': column'
            + ' ' + row[const.COLUMN_NAME] + ' ' + 'has missing value'),

        logs[const.WARNINGS][const.ACTION].append(const.ACTION_MSG)

    return logs

@custom_exceptions.handleError
def check_file_type(df):
    """
    Validated the maven file dataframe as per the basic
    and advanced checks in olmonk.

    Parameters
    ----------
    df : pandas dataframe
        maven output dataframe

    Returns
    -------
    returns the file type according to instrument used
    """

    column_list = df.columns.tolist()
    if set(const.REQUIRED_COL_LIST_MSMS).issubset(set(column_list)):
        return const.MSMS
    elif set(const.MAVEN_REQUIRED_COL_LIST).issubset(set(column_list)) or \
            set(const.MAV_REQUIRED_COL_LIST).issubset(set(column_list)) or set(const.MAVEN_REQUIRED_COL_LIST_OLD).issubset(set(column_list)):
        return const.MS
    elif set(const.ISOCORRECT_REQUIRED_COL_LIST).issubset(set(column_list)):
        return const.ISOCORRECT
    elif set(const.QUANT_REQUIRED_COL_LIST).issubset(set(column_list)):
        return const.QUANTFIT
    else:
        raise custom_exceptions.InvalidMavenError


def check_maven_df_with_olmonk(path, algo=None):
    """
    Validated the maven file dataframe as per the basic
    and advanced checks in olmonk.

    Parameters
    ----------
    path : string
        path
    Returns
    -------
    returns the corrected dataframe and the logs after validation
    """
    logs = {'errors': [], 'warnings': {'action': [], 'message': []}}
    maven_dict = dict(const.DICT_READ_MAVEN_FILE)
    df = read_file(path)
    if algo == "na_lcms":
        maven_dict = formula_and_label_check_for_corna_lcms(maven_dict)
    df = df.dropna(how='all')
    maven_dict['df'] = df
    column_list = df.columns.tolist()
    file_type = check_file_type(df)
    if file_type == const.MS:
        if 'note' in column_list:
            if 'formula' in column_list:
                numerical_columns = list(set(column_list) - set(const.MAVEN_REQUIRED_COL_LIST))
            else:
                numerical_columns = list(set(column_list) - set(const.MAVEN_REQUIRED_COL_LIST_OLD))
        else:
            numerical_columns = list(set(column_list) - set(const.MAV_REQUIRED_COL_LIST))
            df, neg_logs = replace_neg_with_zero(df, numerical_columns)
            for error_msg in neg_logs:
                logs['warnings']['message'].append(error_msg)
        missing_df = hlp.check_missing(df[numerical_columns[1:]])
        fill_missing(df[numerical_columns[1:]])
        maven_dict[const.FUNCTIONS][const.NUMERICAL][
            const.COLUMN_LIST] = numerical_columns

    else:
        numerical_column = [const.PEAK_AREA_CORRECTED]
        df, neg_logs = replace_neg_with_zero(df, numerical_column)
        for error_msg in neg_logs:
            logs['warnings']['message'].append(error_msg)
        missing_df = pd.DataFrame(columns=[const.ROW_NUMBER_COLUMN, const.COLUMN_NAME,
                                           const.STATE])
        maven_dict[const.FUNCTIONS][const.NUMERICAL][
            const.COLUMN_LIST] = numerical_column
    maven_dict['df'] = df
    try:
        cdv = CDV(maven_dict)
        cdv.validate()
        logs = missing_data_report(missing_df, cdv.dv.logs)
        return cdv.dv.corrected_df, logs
    except Exception as e:
        logs['errors'].append(e.message)
        return pd.DataFrame(), logs


def find_extra_components(maven_comp_list, src_list):
    """
    This function returns the components which are not in the
    database to create the mapping file. Returns the missing
    components.

    Args:
    -----
        maven_comp_list: list
            component names list
        src_list: list
            source list component names
    Return:
    -------
        list of diff component names
    """
    diff = set(maven_comp_list) - set(src_list)
    return list(diff)


def find_missing_components(component_list, current_component_list):
    """
    This function returns the components which are not in the
    database to create the mapping file. Returns the missing
    components in the form required in Quant.

    Args:
    -----
        component_list: list
            component names list

    Return:
    -------
        list of dicts of missing component names
    """
    output_dict_list = []
    src_list = current_component_list
    diff  = list(set(component_list) - set(src_list))
    for compound in diff:
        temp_dict = {const.COMPONENT_NAME: compound,
                     const.UNLABELED_FRAGMENT: ""}
        output_dict_list.append(temp_dict)

    return output_dict_list


def find_missing_components_na(component_list, current_component_list):
    """
       This function returns the components which are not in the
       database to create the mapping file. Returns the missing
       components according to Isocorrect application.

       Args:
       -----
           component_list: list
               component names list

       Return:
       -------
           list of dicts of missing component names
       """
    output_dict_list = []
    src_list = current_component_list
    diff  = list(set(component_list) - set(src_list))
    for compound in diff:
        temp_dict = {const.COMPONENT_NAME: compound,
                     const.UNLABELED_FRAGMENT: "",
                     const.FORMULA: "",
                     const.ISOTOPIC_TRACER: "",
                     const.PARENT_FORMULA: ""}
        output_dict_list.append(temp_dict)

    return output_dict_list


def is_maven_file(df):
    """
    This function checks if the file is a maven file
    or not.
    Args:
    -----
        df: Dataframe
            input file dataframe
    Return:
    -------
        Boolean if it is maven file or not
    """
    set_required = set(const.MAVEN_REQUIRED_COL_LIST)
    set_df = set(list(df))
    return set_required.issubset(set_df) or set(const.MAV_REQUIRED_COL_LIST).issubset(set_df) or set(const.MAVEN_REQUIRED_COL_LIST_OLD).issubset(set_df)


def is_maven_file_msms(df):
    """
    This function checks if the file is a MSMS maven file
    or not.
    Args:
    -----
        df: Dataframe
            input file dataframe
    Return:
    -------
        Boolean if it is maven file or not
    """
    set_required = set(const.REQUIRED_COL_LIST_MSMS)
    set_df = set(list(df))
    return set_required.issubset(set_df)


def check_df_type(df):
    """
    This function checks if the file is a MSMS file
    or MS file.
    Args:
    -----
        df: Dataframe
            input file dataframe
    Return:
    -------
        String: MS or MSMS
    """
    if const.MASS_INFO in df.columns:
        if len(df[const.MASS_INFO].unique()) == 1:
            return const.MS
        else:
            return const.MSMS
    else:
        return const.MS


def remove_empty_values(data_dict):
    """
    This function removes the components from the
    data which have partial information from the data editor
    and prevents it to be written to the mapped dataframe

    Args:
    ----------
        data_dict: list of dictionaries
    return:
    ------
        filtered_data_dict: list of dictionaries
    """
    for index in range(len(data_dict)-1, -1 , -1):
        for key, value in list(data_dict[index].items()):
            if not value:
                data_dict.pop(index)
                break
            else:
                pass
    return data_dict


def append_row_from_json(df, data_dict):
    """
    This function appends the rows from the data_dicionary
    to the dataframe passed.

    data_dict = [{'component_name': 'fragment_name'}]
    Args:
    -----
        df: Dataframe
            input file dataframe
        data_dict: list of dictionary
            metadata containing mapping data

    Return:
    -------
        Edited dataframe
    """
    filtered_data_dict = remove_empty_values(data_dict)
    new_df = pd.DataFrame(filtered_data_dict)
    appended_df = df.append(new_df)
    appended_df.reset_index(drop=True)
    return appended_df


def replace_neg_with_zero(input_df, cols):
    """
    This function takes in a dataframes and list of columns
    which are to be checked for negative values and replaces them
    with zero and returns the logs for the same.

    Args
    -----
        input_df: dataframe
        cols: list of columns

    Returns
    --------
        input_df: corrected dataframe with replaced values
        logs: logs for the corrected values.

    """
    logs = list()
    error_dict = dict()
    for col in cols:
        rows = input_df.index[input_df[col] < 0].tolist()
        rows = [x+1 for x in rows]
        unique_rows = list(set(rows))
        error_dict[col] = unique_rows
    for col in cols:
        input_df[col] = input_df[col].clip_lower(0)

    for col, index in list(error_dict.items()):
        if len(error_dict[col]):
            msg = """Row Number {} : column [{}] had negative value which is now converted to 0""".format(index, col)
            logs.append(msg)

    return input_df, logs

def write_file(df, path):
    """
    This function checks the file extension and
    uses the suitable method to write the dataframe in
    that file.

    Parameter
    ----------
        df: dataframe
        path: (String) path to the file

    """
    excel = ['.xls', '.xlsx']

    if os.path.splitext(path)[1] in excel:
        df.to_excel(path, index = False)

    elif os.path.splitext(path)[1] == '.csv':
        df.to_csv(path, index = False)

    elif os.path.splitext(path)[1] == '.txt':
        df.to_csv(path, sep = '\t', index = False)

    else:
        raise custom_exceptions.FileExtensionError

def formula_and_label_check_for_corna_lcms(config_dict):
    """
    Adding formula and Isotope label check in the config dictionary functions
    passing to the olmonk
    Args:
    -----
        config_dict: dict()
                    dict containing function to pass in olmonk

    Return:
    -------
        config_dict: dict()
                        updated dictionary
    """
    config_dict['functions'].update(const.CORNA_LCMS_CHECK)
    return config_dict

